/**
 ** regionMapExecCodels.c
 **
 ** Codels called by execution task regionMapExec
 **
 ** Author: 
 ** Date: 
 **
 **/

#include <portLib.h>
#include <math.h>
#include <limits.h>

#include <stdlib.h>
#include <h2timeLib.h>

#include "server/regionMapHeader.h"

#include <libregionMap.h>
#include <libregionMap_gdhe.h>
#include <stereopixelClient.h>

static REGION_MAP*	regionMap;
static RMAP_3D_PROTO_SPACE*	ps;

static FILE*		loadFile = NULL;
static FILE*		saveFile = NULL;

static POSTER_ID		velodynePosterId;
static velodyne3DImage*		velodyneData;

static POSTER_ID		stereopixelPosterId;
static Spix3DImage*		stereopixelData;

static POSTER_ID	pomPosPosterId;

static DATA_IM3D	im3d; /* Built upon velodyne poster */
static int im3dWidth, im3dHeight;
static int		im3dState = 0; 

static rmapMap*		exportedPoster;
static rmapLightMap*	exportedLightPoster;

/* What follows is for the display */
static int		gdheConnection = 0; /* Indicates wheter gdhe is
					       connected or not */ 
static char		gdheCmdBuf[1024];

/**************************************************************
 *
 * Converts a velodyne3DImage into a DATA_IM3D
 *
 **************************************************************/

void
rmap_fillIm3dFromVyn3DImage (velodyne3DImage* velodyneData,
			     DATA_IM3D* im3d)
{
  int		i, j;
  int		nbLines;
  DATA_PT3D*	point3d; 
  velodyne3DPoint* velodynePoint3d;
  /* Copy the poster into the im3d up to the number of lines that actually
     contains data */
  for (i = 0; i < velodyneData->height; i++) {
    for (j = 0; j < velodyneData->width; j++) {
      point3d = data_im3d_pt (im3d, i, j);
      velodynePoint3d = velodyne3DImageAt(velodyneData, i, j);
      if (velodynePoint3d->status == VELODYNE_GOOD_3DPOINT) {
	point3d->state = GOOD_PT3D;
	point3d->coord_1 = velodynePoint3d->coordinates[0];
	point3d->coord_2 = velodynePoint3d->coordinates[1];
	point3d->coord_3 = velodynePoint3d->coordinates[2];
      }
      else
	point3d->state = BAD_PT3D;
    }
  }
  /* Flag the points not acquired on the remaining lines as BAD_PT3D */
  for (i = 0; i < velodyneData->height; i++) {
    for (j = velodyneData->width; j < velodyneData->maxScanWidth; j++) {
      point3d = data_im3d_pt (im3d, i, j);
      point3d->state = BAD_PT3D;
    }
  }
  im3d->header.coordonnees = COORDONNEES_CARTESIENNES;
  im3d->header.capteur = CAPTEUR_LASER1;
  im3d->header.repere = REPERE_CAPTEUR;
  im3d->header.points = POINTS_CALCULES;
  return ;
}

/**************************************************************
 *
 * Converts a Spix3DImage into a DATA_IM3D
 *
 **************************************************************/

void
rmap_fillIm3dFromStereo3DImage (Spix3DImage* stereopixelData,
				DATA_IM3D* im3d)
{
  int		i, j;
  DATA_PT3D*	pt3d;
  Spix3DPixel*	pix3d;

  pix3d = stereopixelData->pixel;
  pt3d = im3d->points_tab;
  for (i = 0; i < stereopixelData->height; i++) {
    for (j = 0; j < stereopixelData->width; j++) {
      if (pix3d->good != 0) {
	pt3d->state = GOOD_PT3D;
	pt3d->coord_1 = pix3d->x;
	pt3d->coord_2 = pix3d->y;
	pt3d->coord_3 = pix3d->z;
      } 
      else
	pt3d->state = BAD_PT3D;
      pt3d->normales = NIL_NORMALES;
      pt3d++;
      pix3d++;
    }
  }
  im3d->header.points = POINTS_CALCULES;
  im3d->header.repere = REPERE_CAPTEUR;
  return ;
}

/*------------------------------------------------------------------------
 *
 * regionMapExecInit  --  Initialization codel (fIDS, ...)
 *
 * Description: simply sets the SDI values to default values
 * 
 * Returns:    OK or ERROR
 */

STATUS
regionMapExecInit(int *report)
{
  SDI_F->initParams.scale = RMAP_DFLT_SCALE;
  SDI_F->initParams.nbX = RMAP_DFLT_NB_X;
  SDI_F->initParams.nbY = RMAP_DFLT_NB_Y;
  SDI_F->initParams.xOrigin = RMAP_DFLT_X0;
  SDI_F->initParams.yOrigin = RMAP_DFLT_Y0;	
  SDI_F->initParams.robotRadius = RMAP_DFLT_ROBOT_RADIUS;
  SDI_F->initParams.fusionType = RMAP_DFLT_FUSION_TYPE;
  strncpy (SDI_F->initParams.protosFileName, 
	   RMAP_DFLT_PROTOS_FILENAME, 
	   RMAP_MAX_STRLEN);
  strncpy (SDI_F->initParams.pomPosPoster, 
	   RMAP_DFLT_POM_POSTER, 
	   RMAP_MAX_STRLEN);

  SDI_F->connectionParams.connection = RMAP_NO_CONNECTION;
  strncpy (SDI_F->connectionParams.stereopixelPoster, 
	   RMAP_DFLT_STEREOPIXEL_POSTER, 
	   RMAP_MAX_STRLEN);
  strncpy (SDI_F->connectionParams.velodynePoster, 
	   RMAP_DFLT_VELODYNE_POSTER, 
	   RMAP_MAX_STRLEN);

  strncpy (SDI_F->saveParams.fileName, RMAP_DFLT_SAVEFILE, 
	   RMAP_MAX_STRLEN);
  strncpy (SDI_F->loadParams.fileName, RMAP_DFLT_LOADFILE, 
	   RMAP_MAX_STRLEN);
  SDI_F->saveParams.number = RMAP_DFLT_SAVENUM;

/* (commented our for now: it takes too much memory...) */
/*   exportedPoster = posterAddr (REGIONMAP_RMAP_POSTER_ID); */
/*   memset (exportedPoster, 0, sizeof(rmapMap)); */

  exportedLightPoster = posterAddr (REGIONMAP_LIGHTRMAP_POSTER_ID);
  memset (exportedLightPoster, 0, sizeof(rmapLightMap));

  /* Look up for POM position */
  if (regionMapPOM_POSPosterFind(SDI_F->initParams.pomPosPoster, &pomPosPosterId) == ERROR) {
    fprintf(stderr, "regionMap: cannot find pom position poster\n");
    return ERROR;
  }

  regionMap = rMap_createMap ();
  if (regionMap == NULL) {
    fprintf (stderr, "regionMap: could not allocate the internal map structure\n");
    return ERROR;
  }

  return OK;
}

/*------------------------------------------------------------------------
 *
 * regionMapExecEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
regionMapExecEnd(void)
{
  return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description: creates the internal map structure, read the protypes file
 *
 * Reports:      OK
 *	S_regionMap_BAD_INIT_PARAMS
 *	S_regionMap_CAN_NOT_ALLOCATE_MAP
 *	S_regionMap_CAN_NOT_ALLOCATE_PROTO_STRUCTURE
 *	S_regionMap_CAN_NOT_OPEN_FILE
 *	S_regionMap_CAN_NOT_READ_PROTO_FILE
 */

/* regionMapInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapInitMain(rmapInitParams *initParams, int *report)
{
  FILE*	file;

  if (rMap_initMap (regionMap, initParams->scale, 
		    initParams->nbX, initParams->nbY,
		    initParams->xOrigin, initParams->yOrigin, 0.0,
		    initParams->robotRadius) != RMAP_OK) {
    fprintf (stderr, "regionMap: could not allocate the internal map structure\n");
    *report = S_regionMap_CAN_NOT_ALLOCATE_MAP;
    return ETHER;
  }

  ps = rMap_create3dProtoSpace ();
  if (ps == NULL) {
    fprintf (stderr, "regionMap: could not allocate the prototypes structure\n");
    *report = S_regionMap_CAN_NOT_ALLOCATE_PROTO_STRUCTURE;
    return ETHER;
  }
  file = fopen (initParams->protosFileName, "r");
  if (file == NULL) {
    fprintf (stderr, "regionMap: could not open the prototypes file\n");
    *report = S_regionMap_CAN_NOT_OPEN_FILE;
    return ETHER;


  }
  if (rMap_read3DProtoSpace (file, ps) != RMAP_OK) {
    fprintf (stderr, "regionMap: could not read the prototypes file\n");
    *report = S_regionMap_CAN_NOT_READ_PROTO_FILE;
    return ETHER;
  }

  return ETHER;
}

/*------------------------------------------------------------------------
 * Connect
 *
 * Description: 
 *
 * Reports:      OK
 *	S_regionMap_CAN_NOT_CONNECT
 *	S_regionMap_POSTER_NOT_FOUND
 *	S_regionMap_CAN_NOT_CREATE_IM3D
 * 
 */

/* regionMapConnectMain  -  codel EXEC of Connect
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapConnectMain(rmapConnectionParams *connection, int *report)
{
  switch (connection->connection) {
  case RMAP_NO_CONNECTION:
    break;
  case RMAP_STEREOPIXEL_CONNECTION:
  	if (posterFind (connection->stereopixelPoster,&stereopixelPosterId) != OK)
	{
		fprintf (stderr, "regionMap: could not find stereopixelThreeDImage poster \n");
		*report = S_regionMap_CAN_NOT_CONNECT;
		return ETHER;
	}
	/*
	 // We Can not use regionMapSpix3DImagePosterFind as it always fails.
	 // This is due to the fact that static and dynamic sizes are never equal.
    if (regionMapSpix3DImagePosterFind (connection->stereopixelPoster, 
					&stereopixelPosterId) != OK) {
      fprintf (stderr, "regionMap: could not find stereopixelThreeDImage poster \n");
      *report = S_regionMap_CAN_NOT_CONNECT;
      return ETHER;
    } */
    fprintf (stderr, "regionMap: found stereopixel poster %s\n", 
	     connection->stereopixelPoster);
    /* Get the poster address */
    stereopixelData = posterAddr (stereopixelPosterId);
    if (stereopixelData == NULL) {
      *report = S_regionMap_POSTER_NOT_FOUND;
      return ETHER;
    }
    /* Init the im3d */
    im3dWidth = stereopixelData->width;
    im3dHeight = stereopixelData->height;
    if (im3dState != 0) {
      empty_data_im3d (&im3d);
      im3dState = 0;
    }
    if (im3dState == 0) {
      if (init_data_im3d (&im3d, im3dHeight, im3dWidth, 0, 0, 0) 
	  != ERR_IM3D_OK) {
	*report = S_regionMap_CAN_NOT_CREATE_IM3D;
	return ETHER;
      }
      im3dState = 1;
    }
    im3d.header.capteur = CAPTEUR_STEREO;
    im3d.header.coordonnees = COORDONNEES_CARTESIENNES;
    fprintf(stderr, "regionMap: uses %dx%d 3D images from stereopixel\n", 
	    im3dWidth, im3dHeight);
    break;
  case RMAP_VELODYNE_CONNECTION:
    if (regionMapvelodyne3DImagePosterFind (connection->velodynePoster, 
					    &velodynePosterId) != OK) {
      fprintf (stderr, "regionMap: could not find velodyneThreeDImage poster \n");
      *report = S_regionMap_CAN_NOT_CONNECT;
      return ETHER;
    }
    fprintf(stderr, "regionMap: found velodyne poster %s\n", 
	    connection->velodynePoster);
    /* Get the poster address */
    velodyneData = posterAddr (velodynePosterId);
    if (velodyneData == NULL) {
      *report = S_regionMap_POSTER_NOT_FOUND;
      return ETHER;
    }
    /* Init the im3d */
    im3dWidth = velodyneData->maxScanWidth;
    im3dHeight = velodyneData->height;
    if (im3dState != 0) {
      empty_data_im3d (&im3d);
      im3dState = 0;
    }
    if (im3dState == 0) {
      if (init_data_im3d (&im3d, im3dHeight, im3dWidth, 0, 0, 0) 
	  != ERR_IM3D_OK) {
	*report = S_regionMap_CAN_NOT_CREATE_IM3D;
	return ETHER;
      }
      im3dState = 1;
    }
    im3d.header.capteur = CAPTEUR_LASER1;
    im3d.header.coordonnees = COORDONNEES_CARTESIENNES;
    fprintf(stderr, "regionMap: uses %dx%d 3D images from velodyne\n", 
	    im3dWidth, im3dHeight);
    
    break;
  default:
    fprintf (stderr, "regionMap: connection = %d is an impossible case ", 
	     connection->connection);
    break;
  }
  
  return ETHER;
}

/*------------------------------------------------------------------------
 * LoadMap
 *
 * Description: loads a map saved in a file
 *
 * Reports:      OK
 *              S_regionMap_CAN_NOT_LOAD_FILE
 */

/* regionMapLoadMapStart  -  codel START of LoadMap
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadMapStart(rmapLoadParams *loadParams, int *report)
{
  loadFile = fopen (loadParams->fileName, "r");
  if (loadFile == NULL) {
    fprintf (stderr, "regionMap: could not open file %s\n", loadParams->fileName);
    *report = S_regionMap_CAN_NOT_LOAD_FILE;
    return ETHER;
  }
  return EXEC;
}

/* regionMapLoadMapMain  -  codel EXEC of LoadMap
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadMapMain(rmapLoadParams *loadParams, int *report)
{
  if (regionMap != NULL) {
    rMap_destroyMap (regionMap);
    regionMap = rMap_createMap ();
  }
  if (rMap_readAsciiMap (loadFile, regionMap) != RMAP_OK) {
    fprintf (stderr, "regionMap: could load file %s\n", loadParams->fileName);
    *report = S_regionMap_CAN_NOT_LOAD_FILE;
    return ETHER;
  }
  rMap_computeGrowProbas ( regionMap );
  return END;
}

/* regionMapLoadMapEnd  -  codel END of LoadMap
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadMapEnd(rmapLoadParams *loadParams, int *report)
{
  fclose (loadFile);
  loadFile = NULL;
  return ETHER;
}

/*------------------------------------------------------------------------
 * LoadPgmMap
 *
 * Description: Loads an existing map from a PGM file
 *
 * Reports:      OK
 *              S_regionMap_CAN_NOT_LOAD_FILE
 */

/* regionMapLoadPgmMapStart  -  codel START of LoadPgmMap
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadPgmMapStart(rmapLoadPgmParams *loadParams, int *report)
{
  loadFile = fopen (loadParams->fileName, "r");
  if (loadFile == NULL) {
    fprintf (stderr, "regionMap: could not open file %s\n", loadParams->fileName);
    *report = S_regionMap_CAN_NOT_LOAD_FILE;
    return ETHER;
  }
  return EXEC;
}

/* regionMapLoadPgmMapMain  -  codel EXEC of LoadPgmMap
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadPgmMapMain(rmapLoadPgmParams *loadParams, int *report)
{
  if (regionMap != NULL) {
    rMap_destroyMap (regionMap);
    regionMap = rMap_createMap ();
  }
  if (rMap_readPgmFile (loadFile, regionMap,
			loadParams->scale,
			loadParams->xOrigin, loadParams->yOrigin, 0.0,
			loadParams->robotRadius) != RMAP_OK) {
    fprintf (stderr, "regionMap: could load file %s\n", loadParams->fileName);
    *report = S_regionMap_CAN_NOT_LOAD_FILE;
    return ETHER;
  }
  return END;
}

/* regionMapLoadMapEnd  -  codel END of LoadMap
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadPgmMapEnd(rmapLoadPgmParams *loadParams, int *report)
{
  fclose (loadFile);
  loadFile = NULL;
  return ETHER;
}

/*------------------------------------------------------------------------
 * LoadRessacMap
 *
 * Description: loads a map built by Ressac saved in a file
 *
 * Reports:      OK
 *              S_regionMap_CAN_NOT_LOAD_FILE
 *              S_regionMap_NO_MAP_STRUCTURE_INITIALIZED
 */

/* regionMapLoadRessacMapStart  -  codel START of LoadRessacMap
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadRessacMapStart(rmapLoadParams *loadParams, int *report)
{
  loadFile = fopen (loadParams->fileName, "r");
  if (loadFile == NULL) {
    fprintf (stderr, "regionMap: could not open file %s\n", loadParams->fileName);
    *report = S_regionMap_CAN_NOT_LOAD_FILE;
    return ETHER;
  }
  return EXEC;
}

/* regionMapLoadMapRessacMain  -  codel EXEC of LoadRessacMap
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadRessacMapMain(rmapLoadParams *loadParams, int *report)
{
  if (regionMap == NULL) {
    fprintf (stderr, "regionMap: no regionMap structure has been intialized\n");
    *report = S_regionMap_NO_MAP_STRUCTURE_INITIALIZED;
    return ETHER;
  }
  if (rMap_readRessacMap (loadFile, regionMap) != RMAP_OK) {
    fprintf (stderr, "regionMap: could load file %s\n", loadParams->fileName);
    *report = S_regionMap_CAN_NOT_LOAD_FILE;
    return ETHER;
  }
  rMap_computeGrowProbas ( regionMap );
  return END;
}

/* regionMapLoadRessacMapEnd  -  codel END of LoadMap
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapLoadRessacMapEnd(rmapLoadParams *loadParams, int *report)
{
  fclose (loadFile);
  loadFile = NULL;
  return ETHER;
}

/*------------------------------------------------------------------------
 * Save
 *
 * Description: Saves the current internal map in a file
 *
 * Reports:      OK
 *              S_regionMap_RMAP_CAN_NOT_OPEN_FILE
 *              S_regionMap_RMAP_CAN_NOT_SAVE_MAP
 */

/* regionMapSaveStart  -  codel START of Save
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapSaveStart(int *report)
{
  /* Opens the file */
  char fullName[RMAP_MAX_STRLEN];

  sprintf(fullName, "%s.%04d", SDI_F->saveParams.fileName, SDI_F->saveParams.number);
  
  saveFile = fopen(fullName, "w");
  if (saveFile == NULL) {
      fprintf(stderr, "regionMap: cannot open %s\n", fullName);
      *report = S_regionMap_RMAP_CAN_NOT_OPEN_FILE;
      return ETHER;
  }
  return EXEC;
}

/* regionMapSaveMain  -  codel EXEC of Save
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapSaveMain(int *report)
{
  if (rMap_writeAsciiMap (regionMap, saveFile) != RMAP_OK) {
      fprintf(stderr, "regionMap: cannot save map\n");
      *report = S_regionMap_RMAP_CAN_NOT_SAVE_MAP;
      return ETHER;
  }
  return END;
}

/* regionMapSaveEnd  -  codel END of Save
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapSaveEnd(int *report)
{
  fclose(saveFile);
  saveFile = NULL;
  SDI_F->saveParams.number += 1;
  return ETHER;
}

/*------------------------------------------------------------------------
 *
 * Description: Prints info on the current internal map
 *
 * Reports:      OK
 */

/* regionMapPrintMapInfosMain  -  codel Main of PrintMapInfos
   Returns:  ETHER */
ACTIVITY_EVENT
regionMapPrintMapInfosMain(int *report)
{
  rMap_infoMap (regionMap);
}

/*------------------------------------------------------------------------
 * regionMapSetCircularAreaStart
 *
 * Description: Checks that the given radius is non-negative and the
 * probability has a sound value
 *
 * Reports:     OK
 *		S_regionMap_NEGATIVE_RADIUS
 *		S_regionMap_BAD_PROBABILITY_VALUE

 */

/* regionMapSetCircularAreaStart - codel Start of SetCircularArea
   Returns:  ETHER EXEC */
ACTIVITY_EVENT
regionMapSetCircularAreaStart(rmapCircleParams*	circleParams,
			      int*		report)
{
  if (circleParams->radius <= 0.0) {
    fprintf (stderr, "regionMap: circle radius must be > 0\n");
    *report = S_regionMap_NEGATIVE_RADIUS;
    return ETHER;
  }
  if ((circleParams->probaObstacle < 0.0)
      || (circleParams->probaObstacle > 1.0)) {
    fprintf (stderr, "regionMap: the obstacle proba must be >= 0 and <= 1.0\n");
    *report = S_regionMap_BAD_PROBABILITY_VALUE;
    return ETHER;
  }
  else
    return EXEC;
}

/*------------------------------------------------------------------------
 * regionMapSetCircularAreaMain
 *
 * Description: Sets a circular area at the given probability
 *
 * Reports:     OK
 *		S_regionMap_ERROR_IN_LIBRARY
 */
/* regionMapSetCircularAreaMain - codel Main of SetCircularArea
   Returns:  END */
ACTIVITY_EVENT
regionMapSetCircularAreaMain(rmapCircleParams*	circleParams,
			     int*		report)
{
  if (rMap_updateCircularArea (regionMap,
			       circleParams->xCenter,
			       circleParams->yCenter,
			       circleParams->radius,
			       circleParams->probaObstacle) != RMAP_OK) {
    fprintf(stderr, "regionMap: error in call to rMap_updateCircularArea\n");
    *report = S_regionMap_ERROR_IN_LIBRARY;
    return ETHER;
  }
  rMap_computeGrowProbas ( regionMap );

  return END;
}

/*------------------------------------------------------------------------
 * regionMapSetLocalCircularAreaStart
 *
 * Description: Checks that the given radius is non-negative and the
 * probability has a sound value
 *
 * Reports:     OK
 *		S_regionMap_NEGATIVE_RADIUS
 *		S_regionMap_BAD_PROBABILITY_VALUE

 */

/* regionMapSetLocalCircularAreaStart - codel Start of SetLocalCircularArea
   Returns: ETHER EXEC */
ACTIVITY_EVENT
regionMapSetLocalCircularAreaStart(rmapRadiusParams*	circleRadius,
				   int*			report)
{
  if (circleRadius->radius <= 0.0) {
    fprintf (stderr, "regionMap: circle radius must be > 0\n");
    *report = S_regionMap_NEGATIVE_RADIUS;
    return ETHER;
  }
  if ((circleRadius->probaObstacle < 0.0)
      || (circleRadius->probaObstacle > 1.0)) {
    fprintf (stderr, "regionMap: the obstacle proba must be >= 0 and <= 1.0\n");
    *report = S_regionMap_BAD_PROBABILITY_VALUE;
    return ETHER;
  }
  else
    return EXEC;
}

/*------------------------------------------------------------------------
 * regionMapSetLocalCircularAreaMain
 *
 * Description: Sets a circular area centere on the current robot position at
 * the given probability
 *
 * Reports:     OK
 *		S_regionMap_ERROR_IN_LIBRARY
 *		S_regionMap_POM_POSTER_NOT_FOUND
 */
/* regionMapSetLocalCircularAreaMain - codel Main of SetLocalCircularArea
   Returns:  END */
ACTIVITY_EVENT
regionMapSetLocalCircularAreaMain(rmapRadiusParams*	radiusParams,
				  int*			report)
{
  POM_POS	currentPos;

  /* Get the current robot position */
  if (regionMapPOM_POSPosterRead (pomPosPosterId, &currentPos) == ERROR) {
    fprintf(stderr, "regionMap: unable to read pom sensorPos poster\n");
    *report = S_regionMap_POM_POSTER_NOT_FOUND;
    return ETHER;
  }
  
  if (rMap_updateCircularArea (regionMap,
			       currentPos.mainToOrigin.euler.x,
			       currentPos.mainToOrigin.euler.y,
			       radiusParams->radius,
			       radiusParams->probaObstacle) != RMAP_OK) {
    fprintf(stderr, "regionMap: error in call to rMap_updateLocalCircularArea\n");
    *report = S_regionMap_ERROR_IN_LIBRARY;
    return ETHER;
  }

  rMap_computeGrowProbas ( regionMap );

  return END;
}

/*------------------------------------------------------------------------
 * UpdateMap
 *
 * Description: gets the 3D data and fuse them in the regionMap
 *
 * Reports:      OK
 *	S_regionMap_CAN_NOT_READ_POSTER
 *	S_regionMap_NOT_CONNECTED_TO_3D_DATA
 *	S_regionMap_IMAGE3D_ERROR
 *	S_regionMap_FUSION_ERROR
 */

/* regionMapUpdateMapMain  -  codel EXEC of UpdateMap
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapUpdateMapMain(int *report)
{
  T3D		mainToOrigin;
  T3D		sensorToMain;
  T3D		sensorToOrigin;

  t3dInit(&mainToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&sensorToMain, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit(&sensorToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);

  switch (SDI_F->connectionParams.connection) {
  case RMAP_STEREOPIXEL_CONNECTION:
	posterTake(stereopixelPosterId, POSTER_READ);
    /* Convert the stereodata into a DATA_IM3D structure */
    rmap_fillIm3dFromStereo3DImage (stereopixelData, &im3d);
    /* Get the relevant T3D from the PomSensorPos of the stereopixel poster */
    memcpy(&sensorToMain.euler.euler, 
	   &(stereopixelData->imageLeftPos.sensorToMain.euler), sizeof(POM_EULER));
    memcpy(&mainToOrigin.euler.euler, 
	   &(stereopixelData->imageLeftPos.mainToOrigin.euler), sizeof(POM_EULER));
	posterGive(stereopixelPosterId);
    break;
  case RMAP_VELODYNE_CONNECTION:
	posterTake(velodynePosterId, POSTER_READ);
    /* Convert the velodyne data into a DATA_IM3D structure */
    rmap_fillIm3dFromVyn3DImage (velodyneData, &im3d);
    /* Get the relevant T3D from the PomSensorPos of the velodyne poster */
    memcpy(&sensorToMain.euler.euler, 
	   &(velodyneData->position.sensorToMain.euler), sizeof(POM_EULER));
    memcpy(&mainToOrigin.euler.euler, 
	   &(velodyneData->position.mainToOrigin.euler), sizeof(POM_EULER));
	posterGive(velodynePosterId);
    break;
  }

  if (seuille_im3d (&im3d, 0, 0,
			  0, 0, 1, SDI_F->connectionParams.threshold)  != ERR_IM3D_OK) {
	  fprintf(stderr, "regionMap: error in thresholding the 3D image");
	  *report = S_regionMap_IMAGE3D_ERROR;
	  return ETHER;
  }
    
  /* Converts the im3d into the Origin frame */
  t3dCompIn(&sensorToOrigin, &sensorToMain, &mainToOrigin);
  if (change_repere_data_im3d (&im3d, &sensorToOrigin) != ERR_IM3D_OK) {
    fprintf(stderr, "regionMap: error in changing the 3D image frame");
    *report = S_regionMap_IMAGE3D_ERROR;
    return ETHER;
  }

  /* Fuse the 3D image */
  if (rMap_fuseIm3d (regionMap, &im3d, &mainToOrigin ) != RMAP_OK) {
    fprintf(stderr, "regionMap: error while fusing the 3D image");
    *report = S_regionMap_FUSION_ERROR;
    return ETHER;
  }

  /* Now do the classification */
  if (rMap_computeMap3DProbas (regionMap, ps, SDI_F->initParams.fusionType)
      != RMAP_OK) {
    fprintf(stderr, "regionMap: error while fusing the 3D image");
    *report = S_regionMap_FUSION_ERROR;
    return ETHER;
  }
  rMap_computeGrowProbas ( regionMap );

  return ETHER;
}


/*------------------------------------------------------------------------
 * UpdatePoster
 *
 * Description: Updates the poster from the internal map structure
 *
 * Reports:      OK
 *              S_regionMap_CAN_NOT_UPDATE_POSTER
 */

/* regionMapUpdatePosterMain  -  codel EXEC of UpdatePoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */

/* (commented our for now: it takes too much memory...) */

/* ACTIVITY_EVENT */
/* regionMapUpdatePosterMain(int *report) */
/* { */
/*   int		i, j,k; */
/*   RMAP_REGION*	region; */
/*   rmapCell*	cell; */

/*   exportedPoster->scale = regionMap->scale; */
/*   exportedPoster->xNb = regionMap->nblig; */
/*   exportedPoster->yNb = regionMap->nbcol; */
/*   exportedPoster->xOrigin = regionMap->origin2corner.euler.euler[3]; */
/*   exportedPoster->yOrigin = regionMap->origin2corner.euler.euler[4]; */

/*   /\* Loop on the regionMap and the exportedPoster to set the values of the */
/*      poster cells *\/ */
/*   region = regionMap->regions; */
/*   for (i = 0; i < regionMap->nblig; i++) { */
/*     for (j = 0; j < regionMap->nbcol; j++) { */
/*       cell = &(exportedPoster->cells[i*exportedPoster->yNb + j]); */
/*       if (region->infos3d.probaState == RMAP_PROBAS_COMPUTED) { */
/* 	cell->status = RMAP_FILLED; */
/* 	for (k = 0; k < RMAP_NB_CLASSES; k++)  */
/* 	  cell->probas[k] = region->infos3d.growedProbas[k]; */
/*       }  */
/*       else { */
/* 	cell->status = RMAP_EMPTY; */
/* 	for (k = 0; k < RMAP_NB_CLASSES; k++)  */
/* 	  cell->probas[k] = region->infos3d.growedProbas[k]; */
/*       } */
/*     region++; */
/*     } */
/*   } */
  
/*   if (regionMapRmapexportedPosterPosterWrite (REGIONMAP_RMAP_POSTER_ID, exportedPoster) != OK) { */
/*     fprintf (stderr, "regionMap: can not write poster\n"); */
/*     *report = S_regionMap_CAN_NOT_UPDATE_POSTER; */
/*     return ETHER; */
/*   } */

/*   return ETHER; */
/* } */

/*------------------------------------------------------------------------
 * UpdateLightPoster
 *
 * Description: Updates the poster from the internal map structure
 *
 * Reports:     OK
 *              S_regionMap_CAN_NOT_UPDATE_POSTER
 */

/* regionMapUpdateLightPosterMain  -  codel EXEC of UpdateLightPoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapUpdateLightPosterMain(int *report)
{
  int			i,j,k;
  RMAP_REGION*		region;
  unsigned char*	probaObstacle;

  posterTake(REGIONMAP_LIGHTRMAP_POSTER_ID, POSTER_WRITE);
  exportedLightPoster->scale = regionMap->scale;
  exportedLightPoster->xNb = regionMap->nblig;
  exportedLightPoster->yNb = regionMap->nbcol;
  exportedLightPoster->xOrigin = regionMap->origin2corner.euler.euler[3];
  exportedLightPoster->yOrigin = regionMap->origin2corner.euler.euler[4];

  /* Loop on the regionMap and the exportedPoster to set the values of the
     poster cells */
  region = regionMap->regions;
  for (i = 0; i < regionMap->nblig; i++) {
    for (j = 0; j < regionMap->nbcol; j++) {
      probaObstacle = 
	&(exportedLightPoster->probaObstacle[i*exportedLightPoster->yNb + j]);
      *probaObstacle 
	= (unsigned char) round(region->infos3d.growedProbas[RMAP_OBSTACLE]*255.0);
      region++;
    }
  }
  posterGive(REGIONMAP_LIGHTRMAP_POSTER_ID);

  return ETHER;
}

/*------------------------------------------------------------------------
 * SavePoster
 *
 * Description: Saves the poster as a PGM image
 *
 * Reports:      OK
 *              S_regionMap_CAN_NOT_UPDATE_POSTER
 */

/* regionMapUpdatePosterMain  -  codel EXEC of UpdatePoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
regionMapSavePosterMain(int *report)
{
  int		i, j;
  rmapCell*	cell;
  FILE*		file;

  /* Loop on the regionMap and the exportedPoster to set the values of the
     poster cells */

  file = fopen ("/tmp/pipo.pgm", "w");
  if (file == NULL) {
    *report = S_regionMap_CAN_NOT_OPEN_FILE;
    return ETHER;
  }
  fprintf (file, "P2\n%d %d\n255\n", exportedPoster->xNb, exportedPoster->yNb);

  for (i = 0; i < exportedPoster->xNb; i++) {
    for (j = 0; j < exportedPoster->yNb; j++) {
      cell = &(exportedPoster->cells[i*exportedPoster->yNb + j]);
      if (cell->status == RMAP_FILLED) 
	fprintf (file, "%d ", (int) (cell->probas[RMAP_P_OBSTACLE] * 255));
      else 
	fprintf (file, "127 ");
    }
    fprintf (file, "\n");
  }
  
  fclose (file);
  
  return ETHER;
}

/*------------------------------------------------------------------------
 *
 * regionMapDisplayInit  --  Initialization codel (fIDS, ...)
 *
 * Description:  simply initializes the display params to their default value
 * 
 * Returns:    OK or ERROR
 */

STATUS
regionMapDisplayInit(int *report)
{
  strncpy (SDI_F->displayParams.serverName,
	   RMAP_DEFAULT_DPY_SERVER,
	   RMAP_MAX_STRLEN);
  strncpy (SDI_F->displayParams.fileName,
	   RMAP_DEFAULT_DPY_FILE,
	   RMAP_MAX_STRLEN);
  SDI_F->displayParams.fileNumber = RMAP_DEFAULT_DPY_NUMBER;
  SDI_F->displayParams.mergedProbas = RMAP_DEFAULT_DPY_MERGED_PROBAS;
  SDI_F->displayParams.growedProbas = RMAP_DEFAULT_DPY_GROWED_PROBAS;
  
  return OK;
}

/*------------------------------------------------------------------------
 *
 * regionMapDisplayEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
regionMapDisplayEnd(void)
{
  return OK;
}

/*------------------------------------------------------------------------
 * InitDisplay
 *
 * Description: Initializes a display. Opens a connection with gdhe, save the
 * full internal map in a file understandable by gdhe, and displays it.
 *
 * Reports:      OK 
 *	S_regionMap_NO_CONNECTION_POSSIBLE
 *	S_regionMap_OPEN_FILE_FAILED
 *	S_regionMap_BAD_DISPLAY_PARAMS
 *	S_regionMap_CALIFE_ERROR
 * 
 */

/* regionMapInitDisplayExec  -  codel EXEC de InitDisplay
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT regionMapInitDisplayExec (rmapDisplayParams *displayParams, int *report)
{
  FILE*		file;
  GDHE_MNT*	gdheDtm;
  int	error;
  char	displayFileName[RMAP_MAX_STRLEN];
	
  /* Connect with gdhe */
  if (gdheConnection != 0) {
    // this disconnect will be blocking if the gdhe is closed and restarted again
    //disconnect ();
  }

  if (get_connection(SDI_F->displayParams.serverName) == FALSE) {
    *report = S_regionMap_NO_CONNECTION_POSSIBLE;
    return ETHER;
  }
  gdheConnection = 1;

  /* Allocate and fill the structure readable by gdhe */
  gdheDtm = dtm_createGdheMnt ();
  if (gdheDtm == NULL) {
    fprintf(stderr, "dtm: error in dtm_createGdheMnt");
    *report = S_regionMap_CALIFE_ERROR;
    return ETHER;
  }

  error = rMap_mapToGdheMnt (regionMap,
			     displayParams->mergedProbas,
			     displayParams->growedProbas,
			     gdheDtm);
  if (error != RMAP_OK) {
    fprintf(stderr,"regionMap: error in rMap_mapToGdheMnt");
    *report = S_regionMap_CALIFE_ERROR;
    return ETHER;
  }
  
  /* Open the file and write the gdheDtm */
  sprintf (displayFileName, "%s.%04d",
	   SDI_F->displayParams.fileName,
	   SDI_F->displayParams.fileNumber);
  file = fopen(displayFileName, "w");
  if (file == NULL) {
    fprintf(stderr, "regionMap: cannot open %s\n", displayFileName);
    *report = S_regionMap_OPEN_FILE_FAILED;
    return ETHER;
  }
  error = dtm_writeGdheMnt (gdheDtm, file);
  if (error != DTM_OK) {
    fprintf(stderr, "regionMap: error in dtm_writeGdheMnt");
    *report = S_regionMap_CALIFE_ERROR;
    return ETHER;
  }
  SDI_F->displayParams.fileNumber += 1;
  fclose (file);
  
  dtm_destroyGdheMnt (gdheDtm);
  
  /* Now sends the commands to gdhe */
  sprintf (gdheCmdBuf, "initRegionMapDisplay %s mapModel %d %d %d",
	   displayFileName, 1, 1, 1);
  eval_expression (gdheCmdBuf);
  
  return ETHER;
}

ACTIVITY_EVENT regionMapStopDisplayExec (int *report)
{
  /* Disconnect from gdhe */
  if (gdheConnection != 0)
    disconnect ();
  gdheConnection = 0;
  return ETHER;
}

/*------------------------------------------------------------------------
 * UpdateDisplay
 *
 * Description: updates a display
 *
 * Reports:      OK 
 *              S_regionMap_NOT_CONNECTED
 *              S_regionMap_OPEN_FILE_FAILED
 *              S_regionMap_CALIFE_ERROR
 *
 */

/* regionMapUpdateDisplayExec  -  codel EXEC de UpdateDisplay 
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT regionMapUpdateDisplayExec (int *report)
{
  FILE*		file;
  GDHE_MNT*	gdheDtm;
  int		error;
  char	displayFileName[RMAP_MAX_STRLEN];
  double	zScale = 0.01;
  double	zOffset = 0.0;

  /* Check the connection with gdhe */
  if (gdheConnection == 0) {
    *report = S_regionMap_NOT_CONNECTED;
    return ETHER;
  }

  /* Allocate and fill the structure readable by gdhe */
  gdheDtm = dtm_createGdheMnt ();
  if (gdheDtm == NULL) {
    fprintf(stderr, "regionMap: error in dtm_createGdheMnt");
    *report = S_regionMap_CALIFE_ERROR;
    return ETHER;
  }

  error = rMap_lastMapToGdheMnt (regionMap, 
				 SDI_F->displayParams.mergedProbas,
				 SDI_F->displayParams.growedProbas,
				 gdheDtm);
  if (error != RMAP_OK) {
    if (error == ERR_RMAP_NO_CELLS_UPDATED) {
      /* Nothing to do, no cell have been updated inthe map */
      /*       fprintf (stderr, "dtm: nothing to update in the display\n"); */
      return ETHER;
    }
    fprintf(stderr, "regionMap: error in dtm_lastMapToGdheMnt");
    *report = S_regionMap_CALIFE_ERROR;
    return ETHER;
  }
  
  /* Open the file and write the gdheDtm */
  sprintf (displayFileName, "%s.chunk.%04d",
	   SDI_F->displayParams.fileName,
	   SDI_F->displayParams.fileNumber);
  file = fopen(displayFileName, "w");
  if (file == NULL) {
    fprintf(stderr, "regionMap: cannot open %s\n", displayFileName);
    *report = S_regionMap_OPEN_FILE_FAILED;
    return ETHER;
  }
  
  error = dtm_writeGdheMnt (gdheDtm, file);
  if (error != DTM_OK) {
    fprintf(stderr, "regionMap: error in dtm_writeGdheMnt");
    *report = S_regionMap_CALIFE_ERROR;
    return ETHER;
  }
  SDI_F->displayParams.fileNumber += 1;
  fclose (file);
  
  dtm_destroyGdheMnt (gdheDtm);
  
  /* Now sends the commands to gdhe */
  sprintf (gdheCmdBuf, "updateRegionMap mapModel %s", displayFileName);
  eval_expression (gdheCmdBuf);
  
  return ETHER;
}

